mod app;
mod configdata;
mod systemdata;
mod api;
mod utils;
mod clap_ps;

use std::fs;
#[cfg(feature = "avatar")]
use std::fs::remove_file;
use std::path::Path;
use std::collections::HashMap;
#[cfg(feature = "avatar")]
use std::process::Command;

use serde::Serialize;
use serde_json::Value;

use dirs;

use app::*;

use configdata::*;

use systemdata::*;

use clap_ps::*;
use clap::{builder::TypedValueParser, Parser};

use color_eyre::eyre::{eyre, Result};

fn main() -> Result<()> {

    let cli = Args::parse();

    let config_path = match dirs::config_dir() {
        Some(d) => format!("{}/pluralsync", d.display()),
        None => {
            return Err(eyre!("Could not figure out the config directory!"));
        }
    };

    let mut app: App = App::new(config_path.clone());
    app.load_config();
    app.load_system(false);

    match cli.cmd {
        Commands::Sync => {
            app.load_system(true)?;
        },

        Commands::TUI => {
            ()
        },

        Commands::Set { members, #[cfg(feature = "discord")] discord, #[cfg(feature = "fedi")] fedi } => {
            let res = set_member(&app, members);

            #[cfg(feature = "avatar")] {
                match res {
                    Ok(_) => {
                        let _ = update_avatars(config_path.clone(), #[cfg(feature = "discord")] discord, #[cfg(feature = "fedi")] fedi);
                    },
                    Err(_) => (),
                }
            }
        },

        Commands::Add { members, #[cfg(feature = "discord")] discord, #[cfg(feature = "fedi")] fedi } => {
            let res = add_member(config_path.clone(), members);

            #[cfg(feature = "avatar")] {
                match res {
                    Ok(_) => {
                        let _ = update_avatars(config_path.clone(), #[cfg(feature = "discord")] discord, #[cfg(feature = "fedi")] fedi);
                    },
                    Err(_) => (),
                }
            }
        },

        Commands::Get { force_from, #[cfg(feature = "discord")] discord, #[cfg(feature = "fedi")] fedi } => {
            let ff = match force_from {
                Some(x) => x,
                None => ForceFrom::None,
            };

            let _ = get(config_path.clone(), ff);

            #[cfg(feature = "avatar")]
            let _ = update_avatars(config_path.clone(), #[cfg(feature = "discord")] discord, #[cfg(feature = "fedi")] fedi);
        },

        Commands::Members => {
            let _res = memberlist(config_path);
        },
    }
    Ok(())
}

fn set_member(app: &App, tf_members: Vec<String>) -> Result<()> {

    let system = &app.sys.ok_or(eyre!("No system data found"))?;

    let mut to_front: Vec<Member> = Vec::new();
    for member in &tf_members {
        for mem in &system.members {
            let lowcase_aliases: Vec<String> = mem.aliases.iter().map(|x| x.to_lowercase()).collect();
            if mem.name.to_lowercase() == member.to_lowercase() || lowcase_aliases.iter().any(|e| member.contains(e)) {
                println!("Member {member} found.");
                to_front.push(mem.clone());

                break;
            }
        }
    }

    if to_front.len() == tf_members.len() {
        app.set_fronters(to_front)?;
    } BOOKMARK else {
        println!("One or more members were not found. Known members:\n--------------------------");
        let _ = memberlist(app);
        println!("--------------------------\nIf a member is missing from the system try running \"pluralsync sync\" to refresh the local database");
        return Err(eyre!("Missing member"));
    }

    #[cfg(feature = "jlog")] {
        let mut names = Vec::new();
        for m in &to_front {
            names.push(String::from(&m.name));
        }
        let log_fronters = names.join(" || ");

        #[cfg(target_os = "windows")]
        std::process::Command::new("jlog").args(["info", format!("Switch registered: {}", log_fronters)]).output().expect("Logging error");

        #[cfg(not(target_os = "windows"))]
        std::process::Command::new("jlog").arg("info").arg(format!("Switch registered: {}", log_fronters)).output().expect("Logging error");
    }

    Ok(())
}

fn add_member(config_path: String, tf_members: Vec<String>) -> Result<(), &'static str> {
    let config: Config = match get_config(&config_path) {
        Ok(c) => c,
        Err(e) => return Err(e)
    };

    let system: System = get_system(&config_path);

    let mut to_front: Vec<Member> = Vec::new();
    for member in &tf_members {
        for mem in &system.members {
            if mem.name.to_lowercase() == member.to_lowercase() || mem.alias.to_lowercase() == member.to_lowercase() {
                to_front.push(mem.clone());
                break;
            }
        }
    }
    if to_front.len() == tf_members.len() {
        let fronters = get_fronters(&config.pk_key, &config.sp_key, &system, ForceFrom::None);

        let mut aux: Vec<Member> = Vec::new();
        aux.append(&mut fronters.pk.clone());
        aux.append(&mut to_front);
        to_front = aux;

        pk_set_fronters(&config.pk_key, &system, &to_front, &fronters);
        sp_set_fronters(&config.sp_key, &to_front, &fronters);

        let _ = get(config_path, ForceFrom::None);

    } else {
        println!("One or more members were not found. Known members:\n--------------------------");
        let _ = memberlist(config_path);
        println!("--------------------------\nIf a member is missing from the system try running \"pluralsync sync\" to refresh the local database");
        return Err("Missing member");
    }

    #[cfg(feature = "jlog")] {
        let mut names = Vec::new();
        for m in &to_front {
            names.push(String::from(&m.name));
        }
        let log_fronters = names.join(" || ");
        #[cfg(target_os = "windows")]
        std::process::Command::new("jlog").args(["info", format!("Switch registered: {}", log_fronters)]).output().expect("Logging error");
        #[cfg(not(target_os = "windows"))]
        std::process::Command::new("jlog").arg("info").arg(format!("Switch registered: {}", log_fronters)).output().expect("Logging error");
    }

    Ok(())
}

fn memberlist(config_path: String) -> Result<(), &'static str> {
    let sys = get_system(&config_path);

    for mem in sys.members {
        if mem.name != mem.alias {
            println!("{} / {}", mem.name, mem.alias);
        } else {
            println!("{}", mem.name);
        }
    }

    Ok(())

}

fn get(config_path: String, ff: ForceFrom) -> Result<Vec<String>, &'static str> {
    let config: Config = match get_config(&config_path) {
        Ok(c) => c,
        Err(e) => return Err(e)
    };
    let sys = get_system(&config_path);

    let f = get_fronters(&config.pk_key, &config.sp_key, &sys, ff);
    let mut names = Vec::new();
    for m in &f.pk {
        names.push(String::from(&m.name));
    }
    let fronters = names.join(" || ");
    println!("Currently fronting: {}", fronters);
    let _ = fs::write(format!("{}/.front", config_path), fronters);

    Ok(names)
}

#[cfg(feature = "avatar")]
#[allow(unused_variables)]
fn update_avatars(config_path: String, #[cfg(feature = "discord")] discord: bool, #[cfg(feature = "fedi")] fedi: bool) -> Result<(), &'static str>{
    #[allow(unused_mut)]
    let mut config: Config = match get_config(&config_path) {
        Ok(c) => c,
        Err(e) => return Err(e)
    };

    let names = get(config_path, ForceFrom::None).unwrap();


    #[cfg(feature = "avatar")]
    match avatar_module(&config, &names) {
        Ok(_) => {
            #[cfg(feature = "discord")] {
                config.disc_module.enabled |= discord;
                discord_module(&config);
            }
            #[cfg(feature = "fedi")] {
                config.fedi_module.enabled |= fedi;
                fedi_module(&config);
            }
        },
        Err(e) => println!("{}", e),
    }


    Ok(())
}

#[cfg(feature = "avatar")]
fn avatar_module(config: &Config, names: &Vec<String>) -> Result<(), &'static str>{
    if config.avatar_module.enabled {
        let mut whitelisted_names: Vec<String> = names.iter().map(|i| i.to_lowercase()).collect();
        let blacklist: Vec<String> = config.avatar_module.blacklist.iter().map(|i| i.to_lowercase()).collect();
        for name in names {
            if blacklist.contains(&name.to_lowercase()) {
                let index = whitelisted_names.iter().position(|x| *x == name.to_lowercase()).unwrap();
                whitelisted_names.remove(index);
                println!("{} blacklisted from avatar module", name);
            }
        }

        let avatarnames = whitelisted_names.join("").to_lowercase() + ".png";

        if Path::new(&config.avatar_module.avatar_output_path).exists() {
            let _ = remove_file(&config.avatar_module.avatar_output_path);
        }

        if cfg!(target_os = "windows") {
            let cmdaug = format!("copy {}\\{} {}", &config.avatar_module.avatar_folder, avatarnames, &config.avatar_module.avatar_output_path);
            Command::new("cmd").args(["/C", &cmdaug]).output().expect("Avatar module error");
        } else {
            Command::new("sh").arg("-c").arg(format!("cp {}/{} {}", &config.avatar_module.avatar_folder, avatarnames, &config.avatar_module.avatar_output_path)).output().expect("Avatar module error");
        }
        if Path::new(&config.avatar_module.avatar_output_path).exists() {
            println!("Avatar module finished");
            return Ok(())
        } else {
            return Err("Avatar module failed")
        }
    } else {
        Err("Avatar mode disabled")
    }

}

#[cfg(feature = "discord")]
fn discord_module(config: &Config) {
    if config.disc_module.enabled {
        if cfg!(target_os = "windows") {
            let mut c = Command::new("cmd").args(["/C", format!("{} {}", &config.disc_module.python_path, &config.disc_module.script_path).as_str()]).spawn().expect("Discord module error");
            let _ = c.wait().expect("Error");
        } else {
            let mut c = Command::new("sh").arg("-c").arg(format!("{} {}", &config.disc_module.python_path, &config.disc_module.script_path)).spawn().expect("Discord module error");
            let _ = c.wait().expect("Error");
        }
    }
}

#[cfg(feature = "fedi")]
fn fedi_module(config: &Config) {
    if config.fedi_module.enabled {

        let client = reqwest::Client::new();
        let form = Form::new().part("avatar", Part::bytes(fs::read(config.avatar_module.avatar_output_path.clone()).unwrap()).file_name("face.png").mime_str("image/png").unwrap());
        let rb = client
            .patch(config.fedi_module.instance.clone() + "/api/v1/accounts/update_credentials")
            .multipart(form)
            .header(USER_AGENT, "Pluralsync")
            .header(AUTHORIZATION, format!("Bearer {}", &config.fedi_module.token).as_str());

        println!("Fedi module finished");
        match http_request(rb) {
            Ok(_) => (),
            Err(e) => println!("{}", e.to_string()),
        }
    }
}
